# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
import io
import json
import zipfile
from datetime import datetime, timezone
from time import sleep

from persistent import Persistent
from pyramid.config import Configurator
from pyramid.threadlocal import RequestContext, manager
from transaction.interfaces import ITransactionManager
from zope.container.contained import Contained
from zope.interface import Interface
from zope.schema.fieldproperty import FieldProperty

from pyams_scheduler.folder import BaseTaskContainerMixin
from pyams_scheduler.interfaces import AfterRunJobEvent, BeforeRunJobEvent, ITask, SCHEDULER_NAME
from pyams_scheduler.interfaces.task import TASK_STATUS_EMPTY, TASK_STATUS_ERROR, TASK_STATUS_NONE, TASK_STATUS_OK
from pyams_scheduler.interfaces.task.pipeline import IPipelineInput, IPipelineOutput, IPipelineTask, \
    TASK_PIPELINE_INPUT_KEY
from pyams_scheduler.interfaces.task.report import ITaskResultReportInfo
from pyams_scheduler.task import BaseTaskMixin, LOGGER
from pyams_site.interfaces import PYAMS_APPLICATION_DEFAULT_NAME, PYAMS_APPLICATION_SETTINGS_KEY
from pyams_table.testing import OrderedContainer
from pyams_utils.adapter import ContextAdapter, adapter_config, get_annotation_adapter
from pyams_utils.date import get_duration
from pyams_utils.dict import DotDict
from pyams_utils.factory import factory_config
from pyams_utils.registry import get_local_registry, get_pyramid_registry, set_local_registry
from pyams_utils.request import check_request
from pyams_utils.timezone import tztime
from pyams_utils.transaction import COMMITTED_STATUS
from pyams_utils.zodb import ZODBConnection

__docformat__ = 'restructuredtext'

from pyams_scheduler import _


@factory_config(IPipelineInput)
class PipelineInputInfo(Persistent, Contained):
    """Pipeline input persistent info"""
    
    execution_delay = FieldProperty(IPipelineInput['execution_delay'])
    ignore_input_params = FieldProperty(IPipelineInput['ignore_input_params'])
    stop_on_empty_params = FieldProperty(IPipelineInput['stop_on_empty_params'])
    continue_on_empty_result = FieldProperty(IPipelineInput['continue_on_empty_result'])
    
    
@adapter_config(required=ITask,
                provides=IPipelineInput)
def task_pipeline_input(context):
    """Task pipeline input adapter"""
    return get_annotation_adapter(context, TASK_PIPELINE_INPUT_KEY, IPipelineInput)
    
    
class BasePipelineOutput(ContextAdapter):
    """Base pipeline output adapter"""

    def get_values(self, result):
        if not result:
            return {}
        if isinstance(result, str):
            try:
                result = json.loads(result)
            except json.JSONDecodeError:
                result = {}
        if isinstance(result, dict):
            return DotDict(result)
        result = [
            DotDict(item)
            for item in result
        ]
        return result[0] if len(result) == 1 else result
    
    
@factory_config(IPipelineTask)
class PipelineTask(OrderedContainer, BaseTaskContainerMixin, BaseTaskMixin):
    """Pipeline task persistent class"""
    
    label = _("Pipeline task")
    icon_class = 'fas fa-link'

    def __init__(self):
        OrderedContainer.__init__(self)
        BaseTaskMixin.__init__(self)

    def _run(self, report, **kwargs):
        """Task execution wrapper"""
        status = TASK_STATUS_NONE
        result = None
        # initialize ZCA hook
        registry = kwargs.pop('registry', None)
        if registry is None:
            registry = get_pyramid_registry()
        manager.push({'registry': registry, 'request': None})
        config = Configurator(registry=registry)
        config.hook_zca()
        # open ZODB connection
        zodb_connection = ZODBConnection(name=kwargs.pop('zodb_name', ''))
        with zodb_connection as root:
            try:
                application_name = registry.settings.get(PYAMS_APPLICATION_SETTINGS_KEY,
                                                         PYAMS_APPLICATION_DEFAULT_NAME)
                application = root.get(application_name)
                old_registry = get_local_registry()
                try:
                    sm = application.getSiteManager()  # pylint: disable=invalid-name
                    set_local_registry(sm)
                    scheduler_util = sm.get(SCHEDULER_NAME)
                    task = scheduler_util.get_task(self.internal_id)
                    if task is not None:
                        tm = ITransactionManager(task)
                        registry.notify(BeforeRunJobEvent(task))
                        start_date = datetime.now(timezone.utc)
                        request = check_request(base_url=scheduler_util.notified_host,
                                                registry=registry, principal_id=self.principal_id)
                        request.root = application
                        translate = request.localizer.translate
                        task._v_zip_output = zip_output = io.BytesIO()
                        notify = kwargs.pop('notify', True)
                        with zipfile.ZipFile(zip_output, 'w', zipfile.ZIP_DEFLATED) as zip_file:
                            with RequestContext(request):
                                if not (kwargs.pop('run_immediate', False) or task.is_runnable()):
                                    LOGGER.debug(f"Skipping inactive task {task.name}")
                                    return status, result
                                params = kwargs
                                report.writeln(f'Pipeline execution log for task: **{task.name}**', prefix='# ')
                                # loop over internal tasks
                                for inner_task in task.values():
                                    report.writeln('---')
                                    report.writeln(f"Starting task: <strong>{inner_task.name}</strong>",
                                                   prefix='## ')
                                    # check task inputs
                                    task_input = IPipelineInput(inner_task, None)
                                    if task_input is not None:
                                        if task_input.execution_delay:
                                            report.writeln(f'Task execution start delay: '
                                                           f'{task_input.execution_delay} seconds')
                                            sleep(task_input.execution_delay)
                                        if (not params) and task_input.stop_on_empty_params:
                                            report.writeln(f'**Empty params: pipeline execution stopped!**', suffix='\n')
                                            break
                                        if task_input.ignore_input_params:
                                            params = {}
                                    # run internal task
                                    if inner_task.is_zodb_task:
                                        (status, result) = inner_task.run_zodb_task(request, registry,
                                                                                    scheduler_util, report,
                                                                                    notify=False, **params)
                                    else:
                                        (status, result) = inner_task.run_external_task(request, registry,
                                                                                        scheduler_util, report,
                                                                                        notify=False, **params)
                                    if status != TASK_STATUS_OK:
                                        if (status != TASK_STATUS_EMPTY) or \
                                                ((status == TASK_STATUS_EMPTY) and
                                                 (task_input is not None) and
                                                 (not task_input.continue_on_empty_result)):
                                            report.writeln('**Pipeline execution break!**', suffix='\n')
                                            break
                                    # get next task params from execution result
                                    task_output = IPipelineOutput(inner_task, None)
                                    if task_output is not None:
                                        params = {'params': task_output.get_values(result)}
                                    else:
                                        params = {'params': result or {}}
                                    # add task result to internal ZIP
                                    report_info = registry.queryMultiAdapter((inner_task, result),
                                                                             ITaskResultReportInfo)
                                    if report_info is None:
                                        report_info = ITaskResultReportInfo(result, None)
                                    if report_info is not None:
                                        filename = report_info.filename
                                        content = report_info.content
                                        if filename and content:
                                            zip_file.writestr(filename, content)
                                else:
                                    report.writeln('---')
                                    report.writeln('Pipeline execution terminated without error!',
                                                   prefix='### ', suffix='\n')
                        end_date = datetime.now(timezone.utc)
                        duration = (end_date - start_date).total_seconds()
                        report.writeln(f'Total task duration: '
                                       f'**{get_duration(start_date, end_date, request=request)}**', suffix='\n')
                        for attempt in tm.attempts():
                            with attempt as t:
                                if notify:
                                    history_item = task.store_report(result, report, status, start_date, duration)
                                    if status == TASK_STATUS_ERROR:
                                        message_text = translate(_("Task '{}' was executed "
                                                                   "with error")).format(task.name)
                                    else:
                                        message_text = translate(_("Task '{}' was executed "
                                                                   "without error")).format(task.name)
                                    message = task.get_chat_message(request, scheduler_util, status,
                                                                    message_text, history_item)
                                    if message is not None:
                                        message.send()
                                registry.notify(AfterRunJobEvent(task, status, result))
                                if notify:
                                    task.send_report(report, status, history_item, registry)
                                if t.status == COMMITTED_STATUS:
                                    break
                finally:
                    set_local_registry(old_registry)
            except:  # pylint: disable=bare-except
                self._log_exception(None, f"Can't execute scheduled job {self.name}")
            tm = ITransactionManager(self, None)  # pylint: disable=invalid-name
            if tm is not None:
                tm.abort()
        return status, result

    def run(self, report, **kwargs):
        return self._run(report, **kwargs)


@adapter_config(required=(IPipelineTask, Interface),
                provides=ITaskResultReportInfo)
class PipelineTaskReportInfo:
    """Pipeline task report info adapter"""
    
    def __init__(self, task, result=None):
        self.task = task
        
    mimetype = 'application/zip'
    
    @property
    def filename(self):
        now = tztime(datetime.now(timezone.utc))
        return f"report-{now:%Y%m%d}-{now:%H%M%S-%f}.zip"

    @property
    def content(self):
        output = getattr(self.task, '_v_zip_output', None)
        if output is not None:
            output.seek(0)
            return output.read()
        return ''



# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

from zope.interface import Interface
from zope.schema import Bool, Int

from pyams_scheduler.interfaces import ITask
from pyams_scheduler.interfaces.folder import ITaskFolder

__docformat__ = 'restructuredtext'

from pyams_scheduler import _


TASK_PIPELINE_INPUT_KEY = 'pyams_scheduler.task.pipeline.input'


class IPipelineInput(Interface):
    """Pipeline input interface"""
    
    execution_delay = Int(title=_("Execution delay"),
                          description=_("Execution pipeline will wait for the given number of seconds "
                                        "before starting this task"),
                          default=0,
                          min=0)
    
    ignore_input_params = Bool(title=_("Ignore input from previous task"),
                               description=_("If 'yes', output from previous task in the pipeline will be ignored "
                                             "instead of being used as this task input"),
                               required=True,
                               default=False)
    
    stop_on_empty_params = Bool(title=_("Stop pipeline on empty params"),
                                description=_("If 'yes', the pipeline execution will be stopped if params "
                                              "provided by previous task are empty"),
                                required=True,
                                default=False)
    
    continue_on_empty_result = Bool(title=_("Continue pipeline on empty result"),
                                    description=_("If 'yes', the pipeline execution will continue if the "
                                                  "task execution result is empty"),
                                    required=True,
                                    default=False)


class IPipelineOutput(Interface):
    """Pipeline output interface"""
    
    def get_values(self, result):
        """Extract output values mapping from the task execution result
        
        If set, these values will be used as next pipeline task execution parameters.
        """


class IPipelineTaskInfo(Interface):
    """Pipeline task info interface"""


class IPipelineTask(ITaskFolder, ITask, IPipelineTaskInfo):
    """Pipeline task interface"""



# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

import io
import json
import re
import traceback
from datetime import datetime, timezone

from pyams_scheduler.interfaces.task import ITask
from pyams_scheduler.interfaces.task.report import IReport, ITaskResultReportInfo
from pyams_utils.adapter import ContextAdapter, adapter_config
from pyams_utils.factory import factory_config
from pyams_utils.registry import get_pyramid_registry
from pyams_utils.text import text_to_html
from pyams_utils.timezone import tztime

__docformat__ = 'restructuredtext'


@factory_config(IReport)
class Report:
    """Report class helper"""

    def __init__(self, renderer='markdown'):
        self.report = io.StringIO()
        self.renderer = renderer
        self.padding = 0
        
    def add_padding(self):
        self.padding += 1
        
    def remove_padding(self):
        self.padding -= 1
        
    def tell(self):
        return self.report.tell()
    
    def seek(self, position):
        return self.report.seek(position)
    
    def truncate(self):
        return self.report.truncate()
        
    def write(self, value, prefix='', suffix=''):
        if prefix and self.padding:
            prefix = f"{'>' * self.padding} {prefix}"
        if suffix and self.padding:
            suffix = f"{'>' * self.padding} {suffix}"
        return self.report.write(f"{'>' * self.padding}{prefix}{value}{suffix}")
    
    def writeln(self, value, prefix='', suffix=''):
        return self.write(f"{value}\n", prefix, suffix)

    def write_code(self, value):
        value = re.sub('\n*$', '', value, flags=re.MULTILINE)
        return self.writeln(value, prefix='\n```\n', suffix='```\n\n')
    
    def write_shell(self, value):
        return self.write_code(re.sub('^\n', '',
                                      re.sub('\n*$', '', value)))
    
    def write_exception(self, *exc_info):
        return self.write_code('\n'.join(filter(bool,
                                                map(lambda x: re.sub('^\n', '',
                                                                     re.sub('\n*$', '', x,
                                                                            flags=re.MULTILINE)),
                                                    traceback.format_exception(*exc_info)))))
    
    def getvalue(self):
        return text_to_html(self.report.getvalue(), self.renderer)

    def close(self):
        return self.report.close()
    
    
@adapter_config(required=str,
                provides=ITaskResultReportInfo)
class StringTaskResultReportInfo(ContextAdapter):
    """String task result report info adapter"""
    
    json = False
    value = None
    
    def __init__(self, result):
        self.value = result
        try:
            _value = json.loads(result)
        except json.JSONDecodeError:
            pass
        else:
            self.json = True
            
    @property
    def mimetype(self):
        return 'application/json; charset=utf-8' if self.json else 'text/plain; charset=utf-8'
    
    @property
    def filename(self):
        now = tztime(datetime.now(timezone.utc))
        return f"report-{now:%Y%m%d}-{now:%H%M%S-%f}{'.json' if self.json else '.txt'}"

    @property
    def content(self):
        return self.value


@adapter_config(required=dict,
                provides=ITaskResultReportInfo)
class DictTaskResultReportInfo(ContextAdapter):
    """Dict task result report info adapter"""
    
    def __init__(self, result):
        self.result = result
        
    @property
    def mimetype(self):
        return 'application/json; charset=utf-8'
    
    @property
    def filename(self):
        now = tztime(datetime.now(timezone.utc))
        return f"report-{now:%Y%m%d}-{now:%H%M%S-%f}.json"

    @property
    def content(self):
        return json.dumps(self.result)
        

@adapter_config(required=(ITask, list),
                provides=ITaskResultReportInfo)
@adapter_config(required=(ITask, tuple),
                provides=ITaskResultReportInfo)
def list_task_result_report_info(task, result):
    """List task result report info adapter"""
    if not result:
        return None
    result = result[0]
    if not result:
        return None
    registry = get_pyramid_registry()
    adapter = registry.queryMultiAdapter((task, result), ITaskResultReportInfo)
    if adapter is None:
        adapter = ITaskResultReportInfo(result, None)
    return adapter



# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

from zope.container.constraints import contains
from zope.container.interfaces import IContainer
from zope.interface import Interface
from zope.schema import List, Object, TextLine

from pyams_scheduler.interfaces import ITask

__docformat__ = 'restructuredtext'

from pyams_scheduler import _


class ITaskContainer(IContainer):
    """Generic task container interface"""

    contains('pyams_scheduler.interfaces import ITask',
             'pyams_scheduler.interfaces.ITaskContainer')

    folders = List(title=_("Folder sub-folders"),
                   description=_("List of sub-folders assigned to this container"),
                   value_type=Object(schema=Interface),
                   readonly=True)

    tasks = List(title=_("Folder tasks"),
                 description=_("List of tasks assigned to this container"),
                 value_type=Object(schema=ITask),
                 readonly=True)


class ITaskFolder(ITaskContainer):
    """Task folder interface"""

    name = TextLine(title=_("Folder name"),
                    description=_("Descriptive name given to this folder"),
                    required=True)



# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

from pyams_scheduler.interfaces import IScheduler
from pyams_utils.factory import create_object
from pyams_utils.interfaces.monitor import IMonitorExtension, IMonitorExtensionStatus
from pyams_utils.registry import query_utility, utility_config

__docformat__ = 'restructuredtext'


@utility_config(name='pyams_scheduler.monitor',
                provides=IMonitorExtension)
class SchedulerMonitor:
    """Scheduler monitor utility"""
    
    def get_status(self, request):
        scheduler = query_utility(IScheduler)
        if scheduler is None:
            yield create_object(IMonitorExtensionStatus,
                                handler='pyams_scheduler.monitor:scheduler',
                                status='DOWN',
                                message='Missing scheduler utility')
        else:
            yield create_object(IMonitorExtensionStatus,
                                handler='pyams_scheduler.monitor:scheduler',
                                status='UP')
            status_code, message = scheduler.test_process()
            if status_code == 200:
                yield create_object(IMonitorExtensionStatus,
                                    handler='pyams_scheduler.monitor:process',
                                    status='UP',
                                    message=message)
            else:
                yield create_object(IMonitorExtensionStatus,
                                    handler='pyams_scheduler.monitor:process',
                                    status='DOWN',
                                    message=f'Scheduler process not running: {message}')



# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

from pyramid.decorator import reify
from pyramid.view import view_config

from pyams_form.ajax import ajax_form_config
from pyams_form.field import Fields
from pyams_form.interfaces.form import IFormContent, IGroup
from pyams_layer.interfaces import IPyAMSLayer
from pyams_pagelet.pagelet import pagelet_config
from pyams_scheduler.interfaces import ITask, MANAGE_TASKS_PERMISSION
from pyams_scheduler.interfaces.folder import ITaskContainer
from pyams_scheduler.interfaces.task.pipeline import IPipelineInput, IPipelineTask
from pyams_scheduler.task.pipeline import PipelineTask
from pyams_scheduler.task.zmi import BaseTaskAddForm, BaseTaskEditForm
from pyams_scheduler.task.zmi.interfaces import ITaskInnerEditForm
from pyams_scheduler.zmi import ITaskContainerTable, TaskContainerTable, TaskContainerView
from pyams_scheduler.zmi.interfaces import IPipelineTaskContainerTable
from pyams_security.interfaces.base import VIEW_SYSTEM_PERMISSION
from pyams_skin.viewlet.menu import MenuDivider, MenuItem
from pyams_table.interfaces import IColumn
from pyams_utils.adapter import adapter_config
from pyams_utils.factory import factory_config
from pyams_viewlet.viewlet import viewlet_config
from pyams_zmi.form import FormGroupSwitcher
from pyams_zmi.interfaces import IAdminLayer
from pyams_zmi.interfaces.viewlet import IContextAddingsViewletManager
from pyams_zmi.table import ReorderColumn, get_ordered_data_attributes

__docformat__ = 'restructuredtext'

from pyams_scheduler import _


@viewlet_config(name='add-pipeline-task.divider',
                context=ITaskContainer, layer=IAdminLayer, view=TaskContainerTable,
                manager=IContextAddingsViewletManager, weight=899,
                permission=MANAGE_TASKS_PERMISSION)
class PipelineTaskAddMenuDivider(MenuDivider):
    """Pipeline task add menu divider"""
    
    def __new__(cls, context, request, view, manager):
        if IPipelineTask.providedBy(context):
            return None
        return MenuDivider.__new__(cls)


@viewlet_config(name='add-pipeline-task.menu',
                context=ITaskContainer, layer=IAdminLayer, view=TaskContainerTable,
                manager=IContextAddingsViewletManager, weight=900,
                permission=MANAGE_TASKS_PERMISSION)
class PipelineTaskAddMenu(MenuItem):
    """Pipeline task add menu"""
    
    label = _("Add pipeline task...")
    href = 'add-pipeline-task.html'
    modal_target = True
    
    def __new__(cls, context, request, view, manager):
        if IPipelineTask.providedBy(context):
            return None
        return MenuItem.__new__(cls)


class PipelineTaskFormInfo:
    """Pipeline task form info"""
    
    
@ajax_form_config(name='add-pipeline-task.html',
                  context=ITaskContainer, layer=IPyAMSLayer,
                  permission=MANAGE_TASKS_PERMISSION)
class PipelineTaskAddFormInfo(BaseTaskAddForm):
    """Pipeline task add form info"""
    
    content_factory = IPipelineTask
    content_label = PipelineTask.label


@ajax_form_config(name='properties.html',
                  context=IPipelineTask, layer=IPyAMSLayer,
                  permission=MANAGE_TASKS_PERMISSION)
class PipelineTaskEditForm(BaseTaskEditForm):
    """Pipeline task edit form"""
    

@factory_config(IPipelineTaskContainerTable)
class PipelineTaskContainerTable(TaskContainerTable):
    """Pipeline task container table"""
    
    container_class = IPipelineTask
    
    @reify
    def data_attributes(self):
        attributes = super().data_attributes
        container = self.container_class(self.context)
        get_ordered_data_attributes(self, attributes, container, self.request)
        return attributes
    
    
@pagelet_config(name='tasks-list.html',
                context=IPipelineTask, layer=IPyAMSLayer,
                permission=VIEW_SYSTEM_PERMISSION)
class PipelineTaskContainerView(TaskContainerView):
    """Pipeline task container view"""

    table_class = IPipelineTaskContainerTable

    
@adapter_config(name='reorder',
                required=(IPipelineTask, IAdminLayer, ITaskContainerTable),
                provides=IColumn)
class PipelineTaskReorderColumn(ReorderColumn):
    """Pipeline task reorder column"""
    
    
@view_config(name='reorder.json',
             context=IPipelineTask, request_type=IPyAMSLayer,
             renderer='json', xhr=True,
             permission=MANAGE_TASKS_PERMISSION)
def reorder_pipeline_tasks(request):
    """Reorder pipeline tasks"""
    order = request.params.get('order').split(';')
    request.context.updateOrder(order)
    return {
        'status': 'success',
        'closeForm': False
    }


#
# Task pipeline input
#

@adapter_config(name='pipeline-input',
                required=(ITask, IAdminLayer, ITaskInnerEditForm),
                provides=IGroup)
class TaskPipelineInputSubForm(FormGroupSwitcher):
    """Task pipeline input sub form"""
    
    def __new__(cls, context, request, view):
        if not IPipelineTask.providedBy(context.__parent__):
            return None
        return FormGroupSwitcher.__new__(cls)
    
    legend = _("Pipeline input")
    fields = Fields(IPipelineInput)
    
    weight = 10
    
    
@adapter_config(required=(ITask, IAdminLayer, TaskPipelineInputSubForm),
                provides=IFormContent)
def task_pipeline_input_form_content(context, request, view):
    """Pipeline task input form content"""
    return IPipelineInput(context)
    


#
# Copyright (c) 2015-2023 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_scheduler.zmi.folder module

"""

from pyramid.traversal import lineage
from zope.copy import copy
from zope.interface import Interface

from pyams_form.ajax import ajax_form_config
from pyams_form.field import Fields
from pyams_form.interfaces.form import IAJAXFormRenderer
from pyams_layer.interfaces import IPyAMSLayer
from pyams_scheduler.interfaces import IScheduler, MANAGE_SCHEDULER_PERMISSION
from pyams_scheduler.interfaces.folder import ITaskContainer, ITaskFolder
from pyams_scheduler.interfaces.task import ITask
from pyams_scheduler.interfaces.task.pipeline import IPipelineTask
from pyams_scheduler.zmi import TaskContainerTable
from pyams_scheduler.zmi.interfaces import ITaskContainerTable
from pyams_skin.interfaces.viewlet import IBreadcrumbItem, IFormHeaderViewletManager
from pyams_skin.viewlet.help import AlertMessage
from pyams_skin.viewlet.menu import MenuDivider, MenuItem
from pyams_utils.adapter import ContextRequestViewAdapter, adapter_config
from pyams_utils.finder import find_objects_providing
from pyams_utils.intids import get_object_uid
from pyams_utils.traversing import get_parent
from pyams_viewlet.viewlet import viewlet_config
from pyams_zmi.form import AdminModalAddForm, AdminModalEditForm
from pyams_zmi.helper.event import get_json_table_row_add_callback, get_json_table_row_refresh_callback
from pyams_zmi.interfaces import IAdminLayer, IObjectHint, IObjectLabel, IPageTitle, TITLE_SPAN, TITLE_SPAN_BREAK
from pyams_zmi.interfaces.form import IFormTitle
from pyams_zmi.interfaces.table import ITableElementEditor
from pyams_zmi.interfaces.viewlet import IContextAddingsViewletManager
from pyams_zmi.table import TableElementEditor
from pyams_zmi.utils import get_object_label
from pyams_zmi.zmi.viewlet.breadcrumb import AdminLayerBreadcrumbItem

__docformat__ = 'restructuredtext'

from pyams_scheduler import _


@viewlet_config(name='add-folder.menu',
                context=ITaskContainer, layer=IAdminLayer, view=ITaskContainerTable,
                manager=IContextAddingsViewletManager, weight=1,
                permission=MANAGE_SCHEDULER_PERMISSION)
class TaskFolderAddMenu(MenuItem):
    """Task folder add menu"""

    label = _("Add folder...")
    href = 'add-folder.html'
    modal_target = True
    
    def __new__(cls, context, request, view, manager):
        if IPipelineTask.providedBy(context):
            return None
        return MenuItem.__new__(cls)


@ajax_form_config(name='add-folder.html',
                  context=ITaskContainer, layer=IPyAMSLayer,
                  permission=MANAGE_SCHEDULER_PERMISSION)
class TaskFolderAddForm(AdminModalAddForm):
    """Task folder add form"""

    subtitle = _("New folder")
    legend = _("New folder properties")

    fields = Fields(ITaskFolder).select('name')
    content_factory = ITaskFolder

    def add(self, obj):
        self.context[get_object_uid(obj)] = obj


@adapter_config(required=(ITaskContainer, IAdminLayer, TaskFolderAddForm),
                provides=IFormTitle)
def task_folder_add_form_title(context, request, form):
    """Task folder add form title"""
    if IScheduler.providedBy(context):
        return TITLE_SPAN.format(get_object_label(context, request, form))
    scheduler = get_parent(context, IScheduler)
    return TITLE_SPAN_BREAK.format(
        get_object_label(scheduler, request, form),
        get_object_label(context, request, form))


@adapter_config(required=(ITaskContainer, IAdminLayer, TaskFolderAddForm),
                provides=IAJAXFormRenderer)
class TaskFolderAddFormAJAXRenderer(ContextRequestViewAdapter):
    """Task folder add form AJAX renderer"""

    def render(self, changes):
        """AJAX result renderer"""
        if not changes:
            return None
        return {
            'callbacks': [
                get_json_table_row_add_callback(self.context, self.request,
                                                TaskContainerTable, changes)
            ]
        }


@viewlet_config(name='add-folder.divider',
                context=ITaskContainer, layer=IAdminLayer, view=ITaskContainerTable,
                manager=IContextAddingsViewletManager, weight=2,
                permission=MANAGE_SCHEDULER_PERMISSION)
class TaskFolderAddMenuDivider(MenuDivider):
    """Task folder add menu divider"""
    
    def __new__(cls, context, request, view, manager):
        if IPipelineTask.providedBy(context):
            return None
        return MenuDivider.__new__(cls)


@adapter_config(required=(ITaskFolder, IAdminLayer, Interface),
                provides=IObjectHint)
def task_folder_hint(context, request, view):
    """Task folder hint"""
    translate = request.localizer.translate
    return translate(_("Task folder"))


@adapter_config(required=ITaskFolder,
                provides=IObjectLabel)
def task_folder_label(context):
    """Task folder label"""
    return context.name


@adapter_config(required=(ITaskFolder, IAdminLayer, Interface),
                provides=ITableElementEditor)
class TaskFolderElementEditor(TableElementEditor):
    """Task folder element editor"""

    view_name = 'admin'
    modal_target = False


@adapter_config(required=(ITaskFolder, IAdminLayer, Interface),
                provides=IPageTitle)
def task_folder_title(context, request, view):
    """Task folder view title adapter"""
    scheduler = get_parent(context, IScheduler)
    scheduler_label = get_object_label(scheduler, request, view)
    return (f'{scheduler_label}' +
            ''.join((
                f'<small>'
                f'  <small><i class="px-2 fas fa-chevron-right"></i></small> '
                f'  {get_object_label(element, request, view)}'
                f'</small>'
                for element in reversed(list(lineage(context)))
                if ITaskFolder.providedBy(element)))
            )


@adapter_config(required=(ITaskFolder, IAdminLayer, Interface),
                provides=IBreadcrumbItem)
class TaskFolderBreadcrumbItem(AdminLayerBreadcrumbItem):
    """Task folder breadcrumb item"""

    @property
    def label(self):
        return self.context.name


@ajax_form_config(name='properties.html',
                  context=ITaskFolder,
                  permission=MANAGE_SCHEDULER_PERMISSION)
class TaskFolderPropertiesEditForm(AdminModalEditForm):
    """Task folder properties edit form"""

    legend = _("Folder name")

    fields = Fields(ITaskFolder).select('name')


@adapter_config(required=(ITaskFolder, IAdminLayer, TaskFolderPropertiesEditForm),
                provides=IAJAXFormRenderer)
class TaskFolderPropertiesEditFormAJAXRenderer(ContextRequestViewAdapter):
    """Task folder properties edit form AJAX renderer"""

    def render(self, changes):
        if not changes:
            return None
        container = get_parent(self.context, ITaskContainer, allow_context=False)
        return {
            'callbacks': [
                get_json_table_row_refresh_callback(container, self.request,
                                                    TaskContainerTable, self.context)
            ]
        }


@ajax_form_config(name='clone-folder.html',
                  context=ITaskFolder,
                  permission=MANAGE_SCHEDULER_PERMISSION)
class TaskFolderCloneForm(AdminModalAddForm):
    """Task folder clone form"""

    legend = _("Clone folder")

    fields = Fields(ITaskFolder).select('name')

    def create(self, data):
        return copy(self.context)

    def add(self, obj):
        self.context.__parent__[get_object_uid(obj)] = obj
        for task in list(find_objects_providing(obj, ITask)):
            name = task.__name__
            parent = task.__parent__
            parent[get_object_uid(task)] = task
            del parent[name]


@viewlet_config(name='clone-folder.help',
                context=ITaskFolder, layer=IAdminLayer, view=TaskFolderCloneForm,
                manager=IFormHeaderViewletManager, weight=10)
class TaskFolderCloneFormHelp(AlertMessage):
    """Task folder clone form help"""

    status = 'warning'
    _message = _("WARNING: this will duplicate the folder and all it's content, including "
                 "tasks and inner folders; copied tasks will be automatically disabled")


@adapter_config(required=(ITaskFolder, IAdminLayer, TaskFolderCloneForm),
                provides=IAJAXFormRenderer)
class TaskFolderCloneFormAJAXRenderer(ContextRequestViewAdapter):
    """Task folder clone form AJAX renderer"""

    def render(self, changes):
        """AJAX result renderer"""
        if changes is None:
            return None
        container = get_parent(self.context, ITaskContainer, allow_context=False)
        return {
            'callbacks': [
                get_json_table_row_add_callback(container, self.request,
                                                TaskContainerTable, changes)
            ]
        }



# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

from zope.interface import Attribute, Interface

__docformat__ = 'restructuredtext'


class IReport(Interface):
    """Report interface"""

    report = Attribute('Internal report stream')
    renderer = Attribute('Report text renderer')
    
    def add_padding(self):
        """Add padding to report"""
        
    def remove_padding(self):
        """Remove padding from report"""
        
    def tell(self):
        """Return report stream position"""
        
    def seek(self, position):
        """Move stream to given position"""
        
    def truncate(self):
        """Truncate report stream"""
        
    def write(self, value, prefix='', suffix=''):
        """Write the given value to report"""
        
    def writeln(self, value, prefix='', suffix=''):
        """Write the given value to report, followed by a line break"""
        
    def write_code(self, value):
        """Write the given value to report, enclosed in ```"""
    
    def write_exception(self, *exc_info):
        """Write exception traceback to report, enclosed in ```"""
        
    def getvalue(self):
        """Get report content as HTML"""

    def close(self):
        """Close the report stream"""


class ITaskResultReportInfo(Interface):
    """Pipeline report info interface"""
    
    mimetype = Attribute("Report MIME type")
    filename = Attribute("Report filename")
    content = Attribute("Report content")



#
# Copyright (c) 2015-2021 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_scheduler.task.rest module

This module defines REST caller task.
"""

import codecs
import json
import mimetypes
import pprint
import sys
from datetime import datetime, timezone
from http import HTTPStatus
from urllib import parse

import chardet
import requests
from requests import RequestException
from zope.schema.fieldproperty import FieldProperty

from pyams_scheduler.interfaces.task import ITask, TASK_STATUS_ERROR, TASK_STATUS_FAIL, TASK_STATUS_OK
from pyams_scheduler.interfaces.task.pipeline import IPipelineOutput
from pyams_scheduler.interfaces.task.report import ITaskResultReportInfo
from pyams_scheduler.interfaces.task.rest import GET_METHOD, IRESTCallerTask, JSON_CONTENT_TYPE
from pyams_scheduler.task import Task
from pyams_scheduler.task.pipeline import BasePipelineOutput
from pyams_security.interfaces.names import UNCHANGED_PASSWORD
from pyams_utils.adapter import ContextAdapter, ContextRequestAdapter, adapter_config
from pyams_utils.dict import format_dict
from pyams_utils.factory import factory_config
from pyams_utils.html import html_to_text
from pyams_utils.registry import get_pyramid_registry
from pyams_utils.text import render_text
from pyams_utils.timezone import tztime

__docformat__ = 'restructuredtext'

from pyams_scheduler import _  # pylint: disable=ungrouped-imports


@factory_config(IRESTCallerTask)
class RESTCallerTask(Task):
    """REST API caller task"""

    label = _("HTTP service")
    icon_class = 'fab fa-battle-net'

    base_url = FieldProperty(IRESTCallerTask['base_url'])
    service = FieldProperty(IRESTCallerTask['service'])
    headers = FieldProperty(IRESTCallerTask['headers'])
    params = FieldProperty(IRESTCallerTask['params'])
    content_type = FieldProperty(IRESTCallerTask['content_type'])
    verify_ssl = FieldProperty(IRESTCallerTask['verify_ssl'])
    ssl_certs = FieldProperty(IRESTCallerTask['ssl_certs'])
    connection_timeout = FieldProperty(IRESTCallerTask['connection_timeout'])
    allow_redirects = FieldProperty(IRESTCallerTask['allow_redirects'])
    ok_status = FieldProperty(IRESTCallerTask['ok_status'])
    use_proxy = FieldProperty(IRESTCallerTask['use_proxy'])
    proxy_server = FieldProperty(IRESTCallerTask['proxy_server'])
    proxy_port = FieldProperty(IRESTCallerTask['proxy_port'])
    proxy_username = FieldProperty(IRESTCallerTask['proxy_username'])
    _proxy_password = FieldProperty(IRESTCallerTask['proxy_password'])
    authenticate = FieldProperty(IRESTCallerTask['authenticate'])
    username = FieldProperty(IRESTCallerTask['username'])
    _password = FieldProperty(IRESTCallerTask['password'])
    use_jwt_authority = FieldProperty(IRESTCallerTask['use_jwt_authority'])
    jwt_authority_url = FieldProperty(IRESTCallerTask['jwt_authority_url'])
    jwt_token_service = FieldProperty(IRESTCallerTask['jwt_token_service'])
    jwt_login_field = FieldProperty(IRESTCallerTask['jwt_login_field'])
    jwt_password_field = FieldProperty(IRESTCallerTask['jwt_password_field'])
    jwt_token_attribute = FieldProperty(IRESTCallerTask['jwt_token_attribute'])
    jwt_use_proxy = FieldProperty(IRESTCallerTask['jwt_use_proxy'])
    use_api_key = FieldProperty(IRESTCallerTask['use_api_key'])
    api_key_header = FieldProperty(IRESTCallerTask['api_key_header'])
    api_key_value = FieldProperty(IRESTCallerTask['api_key_value'])

    @property
    def password(self):
        """Password getter"""
        return self._password

    @password.setter
    def password(self, value):
        """Password setter"""
        if value == UNCHANGED_PASSWORD:
            return
        self._password = value

    @property
    def proxy_password(self):
        """Proxy password getter"""
        return self._proxy_password

    @proxy_password.setter
    def proxy_password(self, value):
        """Proxy password setter"""
        if value == UNCHANGED_PASSWORD:
            return
        self._proxy_password = value

    @property
    def ok_status_list(self):
        """OK status list getter"""
        return map(int, self.ok_status.split(','))

    def get_request_headers(self, **params):
        """Request HTTP headers getter"""
        result = {}
        for header in (self.headers or ()):
            if not header:
                continue
            try:
                name, value = header.split('=', 1)
            except ValueError:
                continue
            else:
                result[name] = self.get_param_value(value, **params)
        return result

    def get_request_params(self, method, params):
        """Request params getter"""
        result = {}
        if method == GET_METHOD:
            result['params'] = params
        else:
            if self.content_type == JSON_CONTENT_TYPE:
                result['json'] = params
            else:
                result['data'] = params
        return result

    def run(self, report, **kwargs):
        # pylint: disable=too-many-locals,too-many-branches,too-many-statements
        # get remote service URL
        method, service = self.service
        # check proxy configuration
        proxies = {}
        if self.use_proxy:
            parsed = parse.urlparse(self.base_url)
            if self.proxy_username:
                proxy_auth = f'{self.proxy_username}:{self.proxy_password}@'
            else:
                proxy_auth = ''
            proxies[parsed.scheme] = f'http://{proxy_auth}{self.proxy_server}:{self.proxy_port}'
        # check custom headers
        headers = self.get_request_headers(**kwargs)
        # check authorizations
        auth = None
        if self.use_jwt_authority:  # JWT authentication
            jwt_method, jwt_service = self.jwt_token_service
            jwt_service = f'{self.jwt_authority_url}{jwt_service}'
            jwt_params = {
                self.jwt_login_field: self.username,
                self.jwt_password_field: self.password
            }
            try:
                jwt_request = requests.request(jwt_method, jwt_service,
                                               params=jwt_params if jwt_method == 'GET' else None,
                                               json=jwt_params if jwt_method != 'GET' else None,
                                               proxies=proxies if self.jwt_use_proxy else None,
                                               timeout=self.connection_timeout,
                                               allow_redirects=False)
            except RequestException:
                report.writeln('**An HTTP error occurred while getting JWT token**', suffix='\n')
                report.write_exception(*sys.exc_info())
                return TASK_STATUS_FAIL, None
            else:
                status_code = jwt_request.status_code
                report.writeln(f'JWT token status code: {status_code}')
                if status_code != requests.codes.ok:  # pylint: disable=no-member
                    report.writeln(f'JWT headers: {format_dict(jwt_request.headers)}')
                    report.writeln(f'JWT params: {format_dict(jwt_params)}')
                    report.writeln(f'JWT report: {jwt_request.text}', suffix='\n')
                    return TASK_STATUS_ERROR, None
                headers['Authorization'] = f'Bearer ' \
                                           f'{jwt_request.json().get(self.jwt_token_attribute)}'
        elif self.authenticate and self.username:  # Basic authentication
            auth = self.username, self.password
        # launch HTTP request
        results = []
        kwargs_params = kwargs.pop('params', {})
        if isinstance(kwargs_params, dict):
            kwargs_params = [kwargs_params]
        status_code = 0
        for input_params in kwargs_params:
            rest_service = (f'{render_text(self.base_url, **input_params)}'
                            f'{render_text(service, **input_params)}')
            report.writeln(f'HTTP service output', prefix='### ')
            report.writeln(f'HTTP service: `{method} {rest_service}`\n\n')
            if self.use_api_key:  # API key authentication
                headers[self.api_key_header] = self.api_key_value
            if headers:
                report.write('Request headers:')
                report.write_code(format_dict(headers))
            # check params
            params = self.params or {}
            if params:
                params = json.loads(render_text(params, **input_params))
                params.update(input_params)
            else:
                params = input_params
            params.update(kwargs)
            if params:
                report.write('Request params:')
                report.write_code(format_dict(params))
            # build HTTP request
            try:
                rest_response = requests.request(method, rest_service,
                                                 auth=auth,
                                                 headers=headers,
                                                 verify=self.ssl_certs or self.verify_ssl,
                                                 proxies=proxies,
                                                 timeout=self.connection_timeout,
                                                 allow_redirects=self.allow_redirects,
                                                 **self.get_request_params(method, params))
            except RequestException:
                report.writeln('**An HTTP error occurred**', suffix='\n')
                report.write_exception(*sys.exc_info())
                return TASK_STATUS_FAIL, None
            else:
                # check request status
                status_code = rest_response.status_code
                try:
                    status_label = HTTPStatus(status_code).phrase
                except ValueError:
                    status_label = 'Unknown status'
                report.writeln(f'Response status code: `{status_code} - {status_label}`', suffix='\n')
                report.write('Response headers:')
                report.write_code(format_dict(rest_response.headers))
                # check request content
                report.writeln('Response content:')
                content_type = self.get_report_mimetype(rest_response)
                if content_type.startswith('application/json'):
                    response = rest_response.json()
                    message = pprint.pformat(response)
                    report.write_code(message)
                elif content_type.startswith('text/html'):
                    message = html_to_text(rest_response.text)
                    report.writeln(message, suffix='\n')
                elif content_type.startswith('text/'):
                    message = rest_response.text
                    report.writeln(message, suffix='\n')
                else:
                    content = rest_response.content
                    if 'charset=' in content_type.lower():
                        charset = content_type.split('=', 1)[1]
                    else:
                        charset = chardet.detect(content).get('encoding') or 'utf-8'
                    message = codecs.decode(content, charset)
                    report.write_code(message)
                results.append(rest_response)
        return (
            TASK_STATUS_OK if status_code in self.ok_status_list else status_code,
            results
        )
        

@adapter_config(required=IRESTCallerTask,
                provides=IPipelineOutput)
class RESTCallerTaskPipelineOutput(ContextAdapter):
    """REST caller task pipeline output"""
    
    def get_values(self, results):
        registry = get_pyramid_registry()
        values = []
        for result in results:
            content_type = result.headers.get('Content-Type', 'text/plain')
            adapter = registry.queryAdapter(self.context, IPipelineOutput,
                                            name=content_type)
            if adapter is None:
                content_type, _ignored = map(str.strip, content_type.split(';', 1))
                adapter = registry.queryAdapter(self.context, IPipelineOutput,
                                                name=content_type)
            if adapter is not None:
                values.append(adapter.get_values(result))
        return values


@adapter_config(name='text/plain',
                required=IRESTCallerTask,
                provides=IPipelineOutput)
class RESTCallerTaskTextPipelineOutput(BasePipelineOutput):
    """REST caller task plain text pipeline output"""

            
@adapter_config(name='application/json',
                required=IRESTCallerTask,
                provides=IPipelineOutput)
class RESTCallerTaskJSONPipelineOutput(BasePipelineOutput):
    """REST caller task JSON pipeline output"""

    def get_values(self, result):
        return super().get_values(result.json())
    

@adapter_config(required=requests.Response,
                provides=ITaskResultReportInfo)
@adapter_config(required=(ITask, requests.Response),
                provides=ITaskResultReportInfo)
class ResponseReportInfo:
    """HTTP response report info"""

    def __init__(self, *args):
        self.result = args[-1]

    @property
    def mimetype(self):
        return self.result.headers.get('Content-Type', 'text/plain')
    
    @property
    def filename(self):
        content_type = self.mimetype
        if ';' in content_type:
            content_type, _encoding = content_type.split(';', 1)
        extension = mimetypes.guess_extension(content_type)
        now = tztime(datetime.now(timezone.utc))
        return f"report-{now:%Y%m%d}-{now:%H%M%S-%f}{extension or '.txt'}"

    @property
    def content(self):
        result = self.result
        content_type = self.mimetype
        if content_type.startswith('application/json'):
            response = json.dumps(result.json())
        elif content_type.startswith('text/'):
            response = result.text
        else:
            content = result.content
            if 'charset=' in content_type.lower():
                charset = content_type.split('=', 1)[1]
            else:
                charset = chardet.detect(content).get('encoding') or 'utf-8'
            response = codecs.decode(content, charset)
        return response
